@@ -30,8 +30,6 @@
 #include <glib/gi18n.h>
 #include <glib-object.h>
 #include <gst/gst.h>
-#include <gst/gconf/gconf.h>
-#include <gst/tag/tag.h>
 #include <profiles/gnome-media-profiles.h>
 #include "sj-extractor.h"
 #include "sj-structures.h"
@@ -40,8 +38,8 @@
 
 /* Properties */
 enum {
-	PROP_0,
-	PROP_PROFILE,
+  PROP_0,
+  PROP_PROFILE,
 };
 
 /* Signals */
@@ -52,10 +50,15 @@ enum {
   LAST_SIGNAL
 };
 
-static guint sje_table_signals[LAST_SIGNAL] = { 0 };
+static guint signals[LAST_SIGNAL] = { 0 };
 
+/* Default profile name */
 #define DEFAULT_AUDIO_PROFILE_NAME "cdlossy"
 
+/* Element names */
+#define CD_SRC "cdparanoiasrc"
+#define FILE_SINK "gnomevfssink"
+
 struct SjExtractorPrivate {
   /** The current audio profile */
   GMAudioProfile *profile;
@@ -64,20 +67,14 @@ struct SjExtractorPrivate {
   /* The gstreamer pipeline elements */
   GstElement *pipeline, *cdparanoia, *queue, *thread, *encoder, *filesink;
   GstFormat track_format;
-  GstPad *source_pad;
   char *device_path;
   int paranoia_mode;
   int track_start;
   int seconds;
   GError *construct_error;
-  /* Variables used by the thread to push signals into the main thread */
-  guint idle_id;
-  GMutex *idle_mutex;
-  GAsyncQueue *idle_queue;
+  guint tick_id;
 };
 
-static void build_pipeline (SjExtractor *extractor);
-
 /*
  * GObject methods
  */
@@ -93,7 +90,7 @@ G_DEFINE_TYPE(SjExtractor, sj_extractor, G_TYPE_OBJECT);
 static void sj_extractor_class_init (SjExtractorClass *klass)
 {
   GObjectClass *object_class;
-  object_class = (GObjectClass *) klass;
+  object_class = (GObjectClass *)klass;
 
   /* GObject */
   object_class->set_property = sj_extractor_set_property;
@@ -102,10 +99,14 @@ static void sj_extractor_class_init (SjExtractorClass *klass)
 
   /* Properties */
   g_object_class_install_property (object_class, PROP_PROFILE,
-				   g_param_spec_pointer("profile", _("GNOME Audio Profile"), _("The GNOME Audio Profile used for encoding audio"), G_PARAM_READWRITE));
+                                   g_param_spec_object ("profile",
+                                                        _("Audio Profile"),
+                                                        _("The GNOME Audio Profile used for encoding audio"),
+                                                        GM_AUDIO_TYPE_PROFILE,
+                                                        G_PARAM_READWRITE));
 							
   /* Signals */
-  sje_table_signals[PROGRESS] =
+  signals[PROGRESS] =
     g_signal_new ("progress",
                   G_TYPE_FROM_CLASS (object_class),
                   G_SIGNAL_RUN_LAST,
@@ -113,7 +114,7 @@ static void sj_extractor_class_init (SjExtractorClass *klass)
                   NULL, NULL,
                   g_cclosure_marshal_VOID__INT,
                   G_TYPE_NONE, 1, G_TYPE_INT);
-  sje_table_signals[COMPLETION] =
+  signals[COMPLETION] =
     g_signal_new ("completion",
                   G_TYPE_FROM_CLASS (object_class),
                   G_SIGNAL_RUN_LAST,
@@ -121,7 +122,7 @@ static void sj_extractor_class_init (SjExtractorClass *klass)
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
-  sje_table_signals[ERROR] =
+  signals[ERROR] =
     g_signal_new ("error",
                   G_TYPE_FROM_CLASS (object_class),
                   G_SIGNAL_RUN_LAST,
@@ -135,20 +136,17 @@ static void sj_extractor_class_init (SjExtractorClass *klass)
 static void sj_extractor_init (SjExtractor *extractor)
 {
   extractor->priv = g_new0 (SjExtractorPrivate, 1);
-  extractor->priv->profile = gm_audio_profile_lookup(DEFAULT_AUDIO_PROFILE_NAME);
+  extractor->priv->profile = gm_audio_profile_lookup (DEFAULT_AUDIO_PROFILE_NAME);
   extractor->priv->rebuild_pipeline = TRUE;
-  extractor->priv->idle_id = 0;
-  extractor->priv->idle_mutex = g_mutex_new ();
-  extractor->priv->idle_queue = g_async_queue_new ();
 }
 
 static void sj_extractor_set_property (GObject *object, guint property_id,
                                        const GValue *value, GParamSpec *pspec)
 {
-  SjExtractorPrivate *priv = (SjExtractorPrivate*)SJ_EXTRACTOR(object)->priv;
+  SjExtractorPrivate *priv = (SjExtractorPrivate*)SJ_EXTRACTOR (object)->priv;
   switch (property_id) {
   case PROP_PROFILE:
-    priv->profile = (GMAudioProfile *) g_value_get_pointer (value);
+    priv->profile = g_value_get_object (value);
     priv->rebuild_pipeline = TRUE;
     break;
   default:
@@ -159,10 +157,10 @@ static void sj_extractor_set_property (GObject *object, guint property_id,
 static void sj_extractor_get_property (GObject *object, guint property_id,
                                        GValue *value, GParamSpec *pspec)
 {
-  SjExtractorPrivate *priv = (SjExtractorPrivate*)SJ_EXTRACTOR(object)->priv;
+  SjExtractorPrivate *priv = (SjExtractorPrivate*)SJ_EXTRACTOR (object)->priv;
   switch (property_id) {
   case PROP_PROFILE:
-    g_value_set_pointer (value, priv->profile);
+    g_value_set_object (value, priv->profile);
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -172,13 +170,18 @@ static void sj_extractor_get_property (GObject *object, guint property_id,
 static void sj_extractor_finalize (GObject *object)
 {
   SjExtractor *extractor = (SjExtractor *)object;
+
   if (extractor->priv->pipeline)
     gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
+
+  if (extractor->priv->tick_id != 0)
+    g_source_remove (extractor->priv->tick_id);
+
   /* TODO: free the members of priv */
-  g_mutex_free (extractor->priv->idle_mutex);
-  g_async_queue_unref (extractor->priv->idle_queue);
   g_free (extractor->priv);
+
   extractor->priv = NULL;
+
   G_OBJECT_CLASS (sj_extractor_parent_class)->finalize (object);
 }
 
@@ -186,63 +189,91 @@ static void sj_extractor_finalize (GObject *object)
  * Private Methods
  */
 
-typedef struct {
-  guint signal;
-  union {
-    GError *error;
-    int progress;
-  } args;
-} SignalData;
-
-static gboolean
-fire_signal(gpointer user_data)
+static void eos_cb (GstBus *bus, GstMessage *message, gpointer user_data)
 {
   SjExtractor *extractor = SJ_EXTRACTOR (user_data);
-  SignalData *data;
-  g_mutex_lock (extractor->priv->idle_mutex);
+  SjExtractorPrivate *priv;
 
-  while ((data = g_async_queue_try_pop (extractor->priv->idle_queue)) != NULL) {
-    switch (data->signal) {
-    case COMPLETION:
-      g_signal_emit (extractor, sje_table_signals [COMPLETION], 0);
-      break;
-    case ERROR:
-      g_signal_emit (extractor, sje_table_signals[ERROR], 0, data->args.error);
-      g_error_free (data->args.error);
-      break;
-    case PROGRESS:
-      g_signal_emit (extractor, sje_table_signals[PROGRESS], 0, data->args.progress);
-      break;
-    default:
-      g_warning ("Unknown signal ID %d", data->signal);
+  priv = (SjExtractorPrivate*)extractor->priv;
+  
+  gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
+
+  g_source_remove (priv->tick_id);
+  priv->tick_id = 0;
+
+  //extractor->priv->rebuild_pipeline = TRUE;
+
+  g_signal_emit (extractor, signals[COMPLETION], 0);
+}
+
+/* Stolen from gst-plugins-good/ext/gconf/gconf.c */
+static GstPad *
+gst_bin_find_unconnected_pad (GstBin * bin, GstPadDirection direction)
+{
+  GstPad *pad = NULL;
+  GList *elements = NULL;
+  const GList *pads = NULL;
+  GstElement *element = NULL;
+
+  GST_OBJECT_LOCK (bin);
+  elements = bin->children;
+  /* traverse all elements looking for unconnected pads */
+  while (elements && pad == NULL) {
+    element = GST_ELEMENT (elements->data);
+    GST_OBJECT_LOCK (element);
+    pads = element->pads;
+    while (pads) {
+      GstPad *testpad = GST_PAD (pads->data);
+
+      /* check if the direction matches */
+      if (GST_PAD_DIRECTION (testpad) == direction) {
+        GST_OBJECT_LOCK (testpad);
+        if (GST_PAD_PEER (testpad) == NULL) {
+          GST_OBJECT_UNLOCK (testpad);
+          /* found it ! */
+          pad = testpad;
+          break;
+        }
+        GST_OBJECT_UNLOCK (testpad);
+      }
+      pads = g_list_next (pads);
     }
-    g_free (data);
+    GST_OBJECT_UNLOCK (element);
+    elements = g_list_next (elements);
+  }
+  GST_OBJECT_UNLOCK (bin);
+
+  return pad;
+}
+
+/* Stolen from gst-plugins-good/ext/gconf/gconf.c */
+static GstElement *
+gst_gconf_render_bin_from_description (const gchar * description)
+{
+  GstElement *bin = NULL;
+  GstPad *pad = NULL;
+  GError *error = NULL;
+  gchar *desc = NULL;
+
+  /* parse the pipeline to a bin */
+  desc = g_strdup_printf ("bin.( %s )", description);
+  bin = GST_ELEMENT (gst_parse_launch (desc, &error));
+  g_free (desc);
+  if (error) {
+    GST_ERROR ("gstgconf: error parsing pipeline %s\n%s\n",
+        description, error->message);
+    g_error_free (error);
+    return NULL;
   }
 
-  extractor->priv->idle_id = 0;
-  g_mutex_unlock (extractor->priv->idle_mutex);
-  return FALSE;
-}
-
-static void
-queue_signal (SjExtractor *extractor, SignalData *data)
-{
-  g_mutex_lock (extractor->priv->idle_mutex);
-  g_async_queue_push (extractor->priv->idle_queue, data);
-  if (extractor->priv->idle_id == 0)
-    extractor->priv->idle_id = g_idle_add (fire_signal, extractor);
-  g_mutex_unlock (extractor->priv->idle_mutex);
-}
-
-static void eos_cb (GstElement *gstelement, SjExtractor *extractor)
-{
-  SignalData *data;
-  g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
-  gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
-  extractor->priv->rebuild_pipeline = TRUE;
-  data = g_new0 (SignalData, 1);
-  data->signal = COMPLETION;
-  queue_signal (extractor, data);
+  /* find pads and ghost them if necessary */
+  if ((pad = gst_bin_find_unconnected_pad (GST_BIN (bin), GST_PAD_SRC))) {
+    gst_element_add_pad (bin, gst_ghost_pad_new ("src", pad));
+  }
+  if ((pad = gst_bin_find_unconnected_pad (GST_BIN (bin), GST_PAD_SINK))) {
+    gst_element_add_pad (bin, gst_ghost_pad_new ("sink", pad));
+  }
+  return bin;
 }
 
 static GstElement* build_encoder (SjExtractor *extractor)
@@ -261,16 +292,18 @@ static GstElement* build_encoder (SjExtractor *extractor)
   return element;
 }
 
-static void error_cb (GstElement *element, GObject *arg, GError *err, gchar *arg2, gpointer user_data)
+static void error_cb (GstBus *bus, GstMessage *message, gpointer user_data)
 {
   SjExtractor *extractor = SJ_EXTRACTOR (user_data);
-  SignalData *data;
+  GError *error = NULL;
+
   /* Make sure the pipeline is not running any more */
   gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
-  data = g_new0 (SignalData, 1);
-  data->signal = ERROR;
-  data->args.error = g_error_copy (err);
-  queue_signal (extractor, data);
+  extractor->priv->rebuild_pipeline = TRUE;
+
+  gst_message_parse_error (message, &error, NULL);
+  g_signal_emit (extractor, signals[ERROR], 0, error);
+  g_error_free (error);
 }
 
 /**
@@ -286,8 +319,8 @@ static gboolean just_say_yes (GstElement *element, gpointer filename, gpointer u
 static void build_pipeline (SjExtractor *extractor)
 {
   SjExtractorPrivate *priv;
+  GstBus *bus;
 
-  g_return_if_fail (extractor != NULL);
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
 
   priv = extractor->priv;
@@ -295,11 +328,14 @@ static void build_pipeline (SjExtractor *extractor)
   if (priv->pipeline != NULL) {
     gst_object_unref (GST_OBJECT (priv->pipeline));
   }
-  priv->pipeline = gst_thread_new ("pipeline");
-  g_signal_connect (G_OBJECT (priv->pipeline), "error", G_CALLBACK (error_cb), extractor);
+  priv->pipeline = gst_pipeline_new ("pipeline");
+  bus = gst_element_get_bus (priv->pipeline);
+  gst_bus_add_signal_watch (bus);
+
+  g_signal_connect (G_OBJECT (bus), "message::error", G_CALLBACK (error_cb), extractor);
 
   /* Read from CD */
-  priv->cdparanoia = gst_element_factory_make ("cdparanoia", "cdparanoia");
+  priv->cdparanoia = gst_element_factory_make (CD_SRC, "cd_src");
   if (priv->cdparanoia == NULL) {
     g_set_error (&priv->construct_error,
                  SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
@@ -307,22 +343,17 @@ static void build_pipeline (SjExtractor *extractor)
     return;
   }
 
-  g_object_set (G_OBJECT (priv->cdparanoia), "location", priv->device_path, NULL);
+  g_object_set (G_OBJECT (priv->cdparanoia), "device", priv->device_path, NULL);
   g_object_set (G_OBJECT (priv->cdparanoia), "paranoia-mode", priv->paranoia_mode, NULL);
 
   /* Get the track format for seeking later */
   priv->track_format = gst_format_get_by_nick ("track");
-  g_assert (priv->track_format != 0); /* TODO: GError */
-  /* Get the source pad for seeking */
-  priv->source_pad = gst_element_get_pad (priv->cdparanoia, "src");
-  g_assert (priv->source_pad); /* TODO: GError */
+  g_assert (priv->track_format != 0);
 
   priv->queue = gst_element_factory_make ("queue", "queue");
   /* Nice big buffers... */
-  g_object_set (priv->queue, "max-size-time", 20 * GST_SECOND, NULL);
+  g_object_set (priv->queue, "max-size-time", 120 * GST_SECOND, NULL);
   
-  priv->thread = gst_thread_new ("encode thread");
-
   /* Encode */
   priv->encoder = build_encoder (extractor);
   if (priv->encoder == NULL) {
@@ -332,10 +363,10 @@ static void build_pipeline (SjExtractor *extractor)
     return;
   }
   /* Connect to the eos so we know when its finished */
-  g_signal_connect (priv->encoder, "eos", G_CALLBACK (eos_cb), extractor);
+  g_signal_connect (bus, "message::eos", G_CALLBACK (eos_cb), extractor);
 
   /* Write to disk */
-  priv->filesink = gst_element_factory_make ("gnomevfssink", "gnomevfssink");
+  priv->filesink = gst_element_factory_make (FILE_SINK, "file_sink");
   if (priv->filesink == NULL) {
     g_set_error (&priv->construct_error,
                  SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
@@ -345,11 +376,7 @@ static void build_pipeline (SjExtractor *extractor)
   g_signal_connect (G_OBJECT (priv->filesink), "allow-overwrite", G_CALLBACK (just_say_yes), extractor);
 
   /* Add the elements to the pipeline */
-  gst_bin_add (GST_BIN (priv->pipeline), priv->cdparanoia);
-  gst_bin_add (GST_BIN (priv->thread), priv->queue);
-  gst_bin_add (GST_BIN (priv->thread), priv->encoder);
-  gst_bin_add (GST_BIN (priv->thread), priv->filesink);
-  gst_bin_add (GST_BIN (priv->pipeline), priv->thread);
+  gst_bin_add_many (GST_BIN (priv->pipeline), priv->cdparanoia, priv->queue, priv->encoder, priv->filesink, NULL);
 
   /* Link it all together */
   if (!gst_element_link_many (priv->cdparanoia, priv->queue, priv->encoder, priv->filesink, NULL)) {
@@ -366,29 +393,24 @@ static gboolean tick_timeout_cb(SjExtractor *extractor)
 {
   gint64 nanos;
   gint secs;
+  GstState state;
   static GstFormat format = GST_FORMAT_TIME;
 
-  g_return_val_if_fail (extractor != NULL, FALSE);
   g_return_val_if_fail (SJ_IS_EXTRACTOR (extractor), FALSE);
 
-  if (gst_element_get_state (extractor->priv->pipeline) != GST_STATE_PLAYING) {
-    extractor->priv->rebuild_pipeline = TRUE;
+  gst_element_get_state (extractor->priv->pipeline, &state, NULL, GST_CLOCK_TIME_NONE);
+  if (state != GST_STATE_PLAYING) {
     return FALSE;
   }
 
-  if (!gst_pad_query (extractor->priv->source_pad, GST_QUERY_POSITION, &format, &nanos)) {
+  if (!gst_element_query_position (extractor->priv->cdparanoia, &format, &nanos)) {
     g_warning (_("Could not get current track position"));
     return TRUE;
   }
 
   secs = nanos / GST_SECOND;
   if (secs != extractor->priv->seconds) {
-    SignalData *data;
-    extractor->priv->seconds = secs;
-    data = g_new0 (SignalData, 1);
-    data->signal = PROGRESS;
-    data->args.progress = secs - extractor->priv->track_start;
-    queue_signal (extractor, data);
+    g_signal_emit (extractor, signals[PROGRESS], 0, secs - extractor->priv->track_start);
   }
 
   return TRUE;
@@ -417,7 +439,6 @@ GError *sj_extractor_get_new_error (SjExtractor *extractor)
 
 void sj_extractor_set_device (SjExtractor *extractor, const char* device)
 {
-  g_return_if_fail (extractor != NULL);
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
   g_return_if_fail (device != NULL);
 
@@ -425,12 +446,11 @@ void sj_extractor_set_device (SjExtractor *extractor, const char* device)
   extractor->priv->device_path = g_strdup (device);
 
   if (extractor->priv->cdparanoia != NULL)
-    g_object_set (G_OBJECT (extractor->priv->cdparanoia), "location", device, NULL);
+    g_object_set (G_OBJECT (extractor->priv->cdparanoia), "device", device, NULL);
 }
 
 void sj_extractor_set_paranoia (SjExtractor *extractor, const int paranoia_mode)
 {
-  g_return_if_fail (extractor != NULL);
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
 
   extractor->priv->paranoia_mode = paranoia_mode;
@@ -440,12 +460,13 @@ void sj_extractor_set_paranoia (SjExtractor *extractor, const int paranoia_mode)
 
 void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *track, const char* url, GError **error)
 {
-  GstEvent *event;
+  SjExtractorPrivate *priv;
   static GstFormat format = GST_FORMAT_TIME;
   gint64 nanos;
-  SjExtractorPrivate *priv;
+  GstIterator *iter;
+  GstTagSetter *tagger;
+  gboolean done;
 
-  g_return_if_fail (extractor != NULL);
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
 
   g_return_if_fail (url != NULL);
@@ -464,99 +485,95 @@ void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *tra
     }
   }
 
-  /* Set extract speed */
-  g_object_set (G_OBJECT (priv->cdparanoia), "read-speed", 0, NULL);
-
   /* Set the output filename */
   gst_element_set_state (priv->filesink, GST_STATE_NULL);
   g_object_set (G_OBJECT (priv->filesink), "location", url, NULL);
 
   /* Set the metadata */
-  {
-    GList *elts = NULL, *elt = NULL;
-    gboolean taggable = FALSE;
-    if (GST_IS_BIN (priv->encoder)) {
-      elts = (GList *) gst_bin_get_list(GST_BIN (priv->encoder));
-    } else {
-      elts = g_list_append (elts, priv->encoder);
-    }
-    
-    for (elt = elts; elt; elt = elt->next) {
-      if (GST_IS_TAG_SETTER (elt->data)) {
-        gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                            GST_TAG_MERGE_REPLACE_ALL,
-                            GST_TAG_TITLE, track->title,
-                            GST_TAG_ARTIST, track->artist,
-                            GST_TAG_TRACK_NUMBER, track->number,
-                            GST_TAG_TRACK_COUNT, track->album->number,
-                            GST_TAG_ALBUM, track->album->title,
-                            GST_TAG_ENCODER, _("Sound Juicer"),
-                            GST_TAG_ENCODER_VERSION, VERSION,
+  iter = gst_bin_iterate_all_by_interface (GST_BIN (priv->pipeline), GST_TYPE_TAG_SETTER);
+  done = FALSE;
+  while (!done) {
+    switch (gst_iterator_next (iter, (gpointer)&tagger)) {
+    case GST_ITERATOR_OK:
+      /* TODO: generate this as a taglist once, and apply it to all elements */
+      gst_tag_setter_add_tags (tagger,
+                               GST_TAG_MERGE_REPLACE_ALL,
+                               GST_TAG_TITLE, track->title,
+                               GST_TAG_ARTIST, track->artist,
+                               GST_TAG_TRACK_NUMBER, track->number,
+                               GST_TAG_TRACK_COUNT, track->album->number,
+                               GST_TAG_ALBUM, track->album->title,
+                               GST_TAG_ENCODER, _("Sound Juicer"),
+                               GST_TAG_ENCODER_VERSION, VERSION,
+                               NULL);
+#if 0
+      if (track->album->album_id != NULL && strcmp (track->album->album_id, "") != 0) {
+        gst_tag_setter_add_tags (GST_TAG_SETTER (elt->data),
+                            GST_TAG_MERGE_APPEND,
+                            GST_TAG_MUSICBRAINZ_ALBUMID, track->album->album_id,
                             NULL);
-        if (track->album->genre != NULL && strcmp (track->album->genre, "") != 0) {
-          gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-                              GST_TAG_GENRE, track->album->genre,
-                              NULL);
-        }
-        if (track->album->release_date) {
-          gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-                              GST_TAG_DATE, g_date_get_julian (track->album->release_date),
-                              NULL);
-        }
-	if (track->album->album_id != NULL && strcmp (track->album->album_id, "") != 0) {
-	  gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-			      GST_TAG_MUSICBRAINZ_ALBUMID, track->album->album_id,
-			      NULL);
-	}
-	if (track->album->artist_id != NULL && strcmp (track->album->artist_id, "") != 0) {
-	  gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-			      GST_TAG_MUSICBRAINZ_ALBUMARTISTID, track->album->artist_id,
-			      NULL);
-	}
-	if (track->artist_id != NULL && strcmp (track->artist_id, "") != 0) {
-	  gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-			      GST_TAG_MUSICBRAINZ_ARTISTID, track->artist_id,
-			      NULL);
-	}
-	if (track->track_id != NULL && strcmp (track->track_id, "") != 0) {
-	  gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-			      GST_TAG_MUSICBRAINZ_TRACKID, track->track_id,
-			      NULL);
-	}
-	if (track->artist_sortname != NULL && strcmp (track->artist_sortname, "") != 0) {
-	  gst_tag_setter_add (GST_TAG_SETTER (elt->data),
-                              GST_TAG_MERGE_APPEND,
-			      GST_TAG_MUSICBRAINZ_SORTNAME, track->artist_sortname,
-			      NULL);
-	}
-        taggable = TRUE;
       }
-    }
-    if (!taggable) {
-      g_warning ("Couldn't find an encoding element with tag support");
+      if (track->album->artist_id != NULL && strcmp (track->album->artist_id, "") != 0) {
+        gst_tag_setter_add_tags (GST_TAG_SETTER (elt->data),
+                            GST_TAG_MERGE_APPEND,
+                            GST_TAG_MUSICBRAINZ_ALBUMARTISTID, track->album->artist_id,
+                            NULL);
+      }
+      if (track->artist_id != NULL && strcmp (track->artist_id, "") != 0) {
+        gst_tag_setter_add_tags (GST_TAG_SETTER (elt->data),
+                            GST_TAG_MERGE_APPEND,
+                            GST_TAG_MUSICBRAINZ_ARTISTID, track->artist_id,
+                            NULL);
+      }
+      if (track->track_id != NULL && strcmp (track->track_id, "") != 0) {
+        gst_tag_setter_add_tags (GST_TAG_SETTER (elt->data),
+                            GST_TAG_MERGE_APPEND,
+                            GST_TAG_MUSICBRAINZ_TRACKID, track->track_id,
+                            NULL);
+      }
+      if (track->artist_sortname != NULL && strcmp (track->artist_sortname, "") != 0) {
+        gst_tag_setter_add_tags (GST_TAG_SETTER (elt->data),
+                            GST_TAG_MERGE_APPEND,
+                            GST_TAG_MUSICBRAINZ_SORTNAME, track->artist_sortname,
+                            NULL);
+      }
+#endif
+      if (track->album->genre != NULL && strcmp (track->album->genre, "") != 0) {
+        gst_tag_setter_add_tags (tagger,
+                                 GST_TAG_MERGE_APPEND,
+                                 GST_TAG_GENRE, track->album->genre,
+                                 NULL);
+      }
+      if (track->album->release_date) {
+        gst_tag_setter_add_tags (tagger,
+                                 GST_TAG_MERGE_APPEND,
+                                 GST_TAG_DATE, track->album->release_date,
+                                 NULL);
+      }
+      gst_object_unref (tagger);
+      break;
+    case GST_ITERATOR_RESYNC:
+      // TODO?
+      gst_iterator_resync (iter);
+      break;
+    case GST_ITERATOR_ERROR:
+      done = TRUE;
+      break;
+    case GST_ITERATOR_DONE:
+      done = TRUE;
+      break;
     }
   }
-		      
-  /* Let's get ready to rumble! */
-  gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
+  gst_iterator_free (iter);
 
   /* Seek to the right track */
-  event = gst_event_new_segment_seek (priv->track_format | GST_SEEK_METHOD_SET | GST_SEEK_FLAG_FLUSH,
-			              track->number-1, track->number);
-  if (!gst_pad_send_event (priv->source_pad, event)) {
-    g_set_error (error,
-                 SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
-                 _("Could not seek to track"));
-    return;
-  }
+  g_object_set (G_OBJECT (priv->cdparanoia), "track", track->number, NULL);
+  
+  /* Let's get ready to rumble! */
+  gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
+  gst_element_get_state (priv->pipeline, NULL, NULL, GST_CLOCK_TIME_NONE);
 
-  if (!gst_pad_query (priv->source_pad, GST_QUERY_POSITION, &format, &nanos)) {
+  if (!gst_element_query_position (priv->cdparanoia, &format, &nanos)) {
     g_warning (_("Could not get track start position"));
     return;
   }
@@ -564,15 +581,17 @@ void sj_extractor_extract_track (SjExtractor *extractor, const TrackDetails *tra
 
   gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
   
-  g_timeout_add (200, (GSourceFunc)tick_timeout_cb, extractor);
+  priv->tick_id = g_timeout_add (250, (GSourceFunc)tick_timeout_cb, extractor);
 }
 
 void sj_extractor_cancel_extract (SjExtractor *extractor)
 {
-  g_return_if_fail (extractor != NULL);
+  GstState state;
+
   g_return_if_fail (SJ_IS_EXTRACTOR (extractor));
 
-  if (gst_element_get_state (extractor->priv->pipeline) != GST_STATE_PLAYING) {
+  gst_element_get_state (extractor->priv->pipeline, &state, NULL, GST_CLOCK_TIME_NONE);
+  if (state != GST_STATE_PLAYING) {
     return;
   }
   gst_element_set_state (extractor->priv->pipeline, GST_STATE_NULL);
@@ -584,7 +603,7 @@ gboolean sj_extractor_supports_encoding (GError **error)
 {
   GstElement *element = NULL;
 
-  element = gst_element_factory_make ("cdparanoia", "cdparanoia");
+  element = gst_element_factory_make (CD_SRC, "test");
   if (element == NULL) {
     g_set_error (error, SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
                  _("The plugin necessary for CD access was not found"));
@@ -592,7 +611,7 @@ gboolean sj_extractor_supports_encoding (GError **error)
   }
   g_object_unref (element);
 
-  element = gst_element_factory_make ("filesink", "filesink");
+  element = gst_element_factory_make (FILE_SINK, "test");
   if (element == NULL) {
     g_set_error (error, SJ_ERROR, SJ_ERROR_INTERNAL_ERROR,
                  _("The plugin necessary for file access was not found"));
@@ -609,6 +628,7 @@ gboolean sj_extractor_supports_profile (GMAudioProfile *profile)
   GError *error = NULL;
   char *pipeline;
   
+  /* TODO: don't need audioconvert here */
   pipeline = g_strdup_printf ("audioconvert ! %s", gm_audio_profile_get_pipeline (profile));
   element = gst_parse_launch (pipeline, &error);
   /* TODO: check error */
